/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.skills.strategies.follow.SimpleFollowStrategy;
import de.unibi.airobots.bonsai.strategies.follow.FollowStrategy;
import de.unibi.airobots.btl.data.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.SpeedData;
import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * In this state the robot follows given person.
 * 
 * @author lkettenb
 */
public class FollowPerson extends AbstractState {
    /*
     * Sensors used by this state.
     */
    private Sensor<PersonList> personSensor;
    private Sensor<PositionData> posSensor;
    private Sensor<SpeedData> speedSensor;
    private Sensor<GlobalPlannerStateChange> gplanChangeSensor;
    
    /*
     * Actuators used by this state.
     */
    private SpeechActuator speechActuator;
    private NavigationActuator navActuator;
    
    /**
     * Time to wait in follow loop (milliseconds).
     */
    private static final int WAIT_FOLLOW_LOOP = 400;
    
    /**
     * The logger.
     */
    private Logger logger = getLogger();
    
    /**
     * The follow strategy to use.
     */
    private FollowStrategy followStrategy;
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public FollowPerson(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }
    
    @Override
    public ExitStatus init() throws InitException {
        // Initialize sensors
        personSensor = getManagedSensor("PersonSensor");
        posSensor = getManagedSensor("PositionSensor");
        speedSensor = getManagedSensor("SpeedSensor");
        gplanChangeSensor = getManagedSensor("GlobalPlannerStateChangeSensor");
        
        // Initialize actuators
        speechActuator = getManagedActuator("SpeechActuator");
        navActuator = getManagedActuator("NavigationActuator");
        
        followStrategy = new SimpleFollowStrategy(
                personSensor, posSensor, speedSensor, speechActuator, 
                navActuator, gplanChangeSensor);
        
        followStrategy.init();
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus execute() throws ExecuteException {
        try {
            followStrategy.execute();
        } catch (IOException ex) {
            // Too bad. Unable to follow any person!
            throw new ExecuteException(ex.getMessage());
        }
       
        try {
            sleep(WAIT_FOLLOW_LOOP);
        } catch (InterruptedException ex) {
            logger.warn(ex.getMessage());
        }
        
        setProcessingStatus(getWaitProcessingMessage());
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }
    
}
