/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.helper.PersonManager;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.btl.data.PersonData;
import java.io.IOException;
import java.util.logging.Level;
import org.apache.log4j.Logger;

/**
 * Use this state to wait until a person is recognized in front of the robot.
 * 
 * 
 * @author lkettenb
 */
public class WaitForPerson extends AbstractState {
    
    /*
     * Sensors used by this state.
     */
    private Sensor<PersonList> personSensor;
    
    /**
     * The logger.
     */
    private Logger logger = getLogger();
    
    /**
     * Time to sleep before requesting new data.
     */
    private static final int TIME_TO_SLEEP = 350;
    
    /**
     * Constructor.
     * 
     * @param aStateMachine The state machine to use.
     */
    public WaitForPerson(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }
    
    @Override
    public ExitStatus init() throws InitException {
        // Initialize sensors
        personSensor = getManagedSensor("PersonSensor");
        
        // Wait for persons in front of sensor
        logger.debug("Waiting for person in front ...");
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus execute() throws ExecuteException {
        PersonList persons;
        
        try {
            persons = personSensor.readLast();
        } catch (IOException ex) {
            throw new ExecuteException(ex.getMessage());
        } catch (InterruptedException ex) {
            throw new ExecuteException(ex.getMessage());
        }
        
        for (PersonData p : persons) {
            if (PersonManager.isStandingInFront(p)) {
                logger.debug("Person in front.");
                return ExitStatus.SUCCESS;
            } 
        }
        
        try {
            sleep(TIME_TO_SLEEP);
        } catch (InterruptedException ex) {
            logger.warn(ex.getMessage());
        }
        
        setProcessingStatus(getWaitProcessingMessage());
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }
}
