/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.strategies.follow;

import java.io.IOException;
import java.util.LinkedList;

import net.sf.xcf.transport.XOPData;
import nu.xom.Builder;
import nu.xom.ParsingException;
import nu.xom.ValidityException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.TaskPatternActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.btl.data.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PersonData;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.SpeedData;
import de.unibi.airobots.btl.data.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.taskpattern.PersonFoundNotification;
import de.unibi.airobots.btl.data.taskpattern.PleaseWaitNotification;
import de.unibi.airobots.btl.data.taskpattern.RobotLostHumanNotification;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.ReachStatus;
import de.unibi.airobots.bonsai.xcf.actuators.XcfTaskPatternActuator;

public class TaskPatternBasedPersonTooFarFollowStrategy implements
        FollowStrategy {

    private Logger logger =
            Logger.getLogger(TaskPatternBasedPersonTooFarFollowStrategy.class);

    private static final double MAX_PERCEPTION_AGE = 2000;

    private static final double DISTANCE_THRESHOLD_STOP = 1000;

    private static final double DISTANCE_THRESHOLD_MOVE = 3000;

    private static final double DISTANCE_THRESHOLD_TURN = 1000;

    private static final double DISTANCE_THRESHOLD_START_FOLLOWING = 3000;

    private static final double ANGLE_THRESHOLD_START_FOLLOWING = Math.PI * 0.25;
    private static final double STEP_FORWARD = 1.0;

    private static final double RESUME_MAX_DIST = 1900;

    private static final double COORDINATE_TOLERANCE = 0.5;
    private static final double YAW_TOLERANCE = Math.PI * 0.3;

    private static final int QUEUE_LENGTH = 10;

    private int personFollowId = -1;

    private int noPersonCounter = 0;

    private LinkedList<NavigationGoalData> lastKnownPositions =
            new LinkedList<NavigationGoalData>();

    private volatile boolean firstTimeLost = true;

    Sensor<PersonList> personSensor;
    Sensor<PositionData> posSensor;
    Sensor<SpeedData> speedSensor;
    NavigationActuator navActuator;
    TaskPatternActuator taskActuator;
    Sensor<GlobalPlannerStateChange> gplanChangeSensor;

    private Builder builder = new Builder();

    public TaskPatternBasedPersonTooFarFollowStrategy() {

        personSensor =
                BonsaiManager.getInstance().createSensor("PersonSensor",
                        PersonList.class);
        posSensor =
                BonsaiManager.getInstance().createSensor("PositionSensor",
                        PositionData.class);
        taskActuator =
                BonsaiManager.getInstance().createActuator(
                        "TaskPatternActuator", XcfTaskPatternActuator.class);
        navActuator =
                BonsaiManager.getInstance().createActuator(
                        "NavigationActuator", NavigationActuator.class);
        speedSensor =
                BonsaiManager.getInstance().createSensor("SpeedSensor",
                        SpeedData.class);
        gplanChangeSensor =
                BonsaiManager.getInstance().createSensor(
                        "GlobalPlannerStateChangeSensor",
                        GlobalPlannerStateChange.class);
    }

    @Override
    public void init() {
        personFollowId = -1;
        lastKnownPositions.clear();
    };

    @Override
    public void execute() throws IOException {

        logger.debug("check for person");

        PersonList personList;
        try {
            personList = personSensor.readLast();
        } catch (InterruptedException e1) {
            logger.fatal(e1.getMessage());
            return;
        }

        PersonData personFollow = null;
        // try to find followed person.
        if (personFollowId == -1) {
            personFollow = searchNewPerson(personList);
        } else {
            personFollow = searchKnownPerson(personList);
        }

        // If a followed person was defined.
        logger.debug("person Follow " + personFollow);
        if (personFollow != null) {
            follow(personFollow);
        }
    }

    @Override
    public void execute(int personId) throws IOException {
        logger.debug("setting personFollowId to " + personId);
        personFollowId = personId;
        execute();
    }

    private PersonData searchNewPerson(PersonList personList)
            throws IOException {

        PersonData personFollow = null;

        // choose nearest person
        // double minDist = Double.MAX_VALUE;
        double minAngle = Math.PI;

        for (PersonData p : personList) {

            boolean isNear =
                    p.getDistance() <= DISTANCE_THRESHOLD_START_FOLLOWING;

            boolean isSmallId = p.getId() >= 0;

            boolean isInFront =
                    Math.abs(p.getAngle()) <= ANGLE_THRESHOLD_START_FOLLOWING;

            if (p.getAngle() < minAngle && isNear && isSmallId && isInFront) {
                //minDist = p.getDistance();
                minAngle = p.getAngle();
                personFollow = p;
            }
        }

        if (personFollow != null) {

            personFollowId = personFollow.getId();
            logger.debug("searchNewPerson: Following person [" + personFollowId + "]");

        } else {

            if (firstTimeLost && !lastKnownPositions.isEmpty()) {
                try {
                    taskActuator.initiate(new XOPData(builder.build("<"
                            + RobotLostHumanNotification.getBaseTag() + "/>",
                            null)));
                } catch (ValidityException e) {
                    logger.error(e);
                } catch (ParsingException e) {
                    logger.error(e);
                }
                firstTimeLost = false;
            }
            logger.debug("No person detected");
            stepForward();
            noPersonCounter++;

        }
        return personFollow;
    }

    private void stepForward() {
        logger.info("Making a step forward.");
        // make a step forward
        PositionData positionData;
        try {
            positionData = posSensor.readLast();

            NavigationGoalData g;
            g =
                    CoordinateSystemConverter.polar2NavigationGoalData(
                            positionData, 0, STEP_FORWARD);

            g.setYawTolerance(YAW_TOLERANCE);
            g.setCoordinateTolerance(COORDINATE_TOLERANCE);
            g.setType(GoalType.GLOBAL);

            GoalReachedWaiter waiter =
                    new GoalReachedWaiter(posSensor, speedSensor,
                            gplanChangeSensor, navActuator, 800,
                            GoalReachedWaiter.DEFAULT_STEPS);
            navActuator.setGoal(g);
            ReachStatus goalReached = waiter.waitForReachingGoal(g, 5000);
            logger.debug("goalReached: " + goalReached);
        } catch (InterruptedException e) {
            logger.warn("Interrupted while waiting for position data.");
        } catch (IOException e) {
            logger.warn("IOException while waiting for position data.");
        }

    }

    private PersonData searchKnownPerson(PersonList personList)
            throws IOException {

        try {
            personList = personSensor.readLast();
        } catch (InterruptedException e1) {
            logger.fatal(e1.getMessage());
        }

        PersonData personFollow = personList.getPersonById(personFollowId);

        logger.info("personData " + personFollow);
        String list = "";
        for (PersonData d : personList) {
            list += " " + d.getId();
        }
        logger.info("searchKnownPerson: following person " + personFollow + ". Found IDs in scan:"
                + list);

        // search for followed person, if not in person list.
        if (personFollow == null) {
            try {
                PositionData currentPosition = posSensor.readLast();

                // choose position
                // TODO check m / mm
                double minDist2 = Double.MAX_VALUE;
                int nearestId = -1;
                for (NavigationGoalData d : lastKnownPositions) {
                    double dist =
                            Math.hypot(currentPosition.getX() - d.getX(),
                                    currentPosition.getY() - d.getY());
                    if (dist < minDist2) {
                        minDist2 = dist;
                        nearestId = lastKnownPositions.indexOf(d);
                    }
                }

                try {
                    taskActuator
                            .initiate(new XOPData(builder.build("<"
                                    + PleaseWaitNotification.getBaseTag()
                                    + "/>", null)));
                } catch (ValidityException e) {
                    logger.error(e);
                } catch (ParsingException e) {
                    logger.error(e);
                }

                NavigationGoalData lastKnownPosition =
                        lastKnownPositions.get(nearestId);

                if (lastKnownPosition != null) {
                    logger.debug("go to lastKnownPosition");
                    lastKnownPosition
                            .setCoordinateTolerance(COORDINATE_TOLERANCE);
                    lastKnownPosition.setYawTolerance(YAW_TOLERANCE);
                    GoalReachedWaiter waiter =
                            new GoalReachedWaiter(posSensor, speedSensor,
                                    gplanChangeSensor, navActuator, 800,
                                    GoalReachedWaiter.DEFAULT_STEPS);

                    navActuator.setGoal(lastKnownPosition);
                    ReachStatus goalReached =
                            waiter
                                    .waitForReachingGoal(lastKnownPosition,
                                            10000);
                    if (goalReached == ReachStatus.REACHED) {
                        logger.debug("last known position reached");
                    } else {
                        logger
                                .warn("---> last known position not reached <--- ("
                                        + goalReached + ")");
                    }
                }
            } catch (IndexOutOfBoundsException e) {
                logger.warn("Desired position id not available.");
            } catch (InterruptedException e) {
                logger.warn(e.getMessage());
            }

            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                logger.warn("Interrupted while waiting");
            }

            try {
                personList = personSensor.readLast();
                personFollow = personList.getPersonById(personFollowId);
                if (personList != null && personFollowId != -1
                        && personFollow != null) {
                    logger.info("found person...");
                    logger.debug("get person " + personList);
                    return personFollow;
                }
                logger.info("looking for new person in front");

                if (personList.size() != 0) {
                    // choose nearest person
                    double minDist = Double.MAX_VALUE;
                    for (PersonData p : personList) {
                        // boolean isYoung = System.currentTimeMillis()
                        // - p.getTimestamp().getUpdated().getTime() <
                        // MAX_PERCEPTION_AGE;
                        boolean isSmallId = p.getId() < 1000 && p.getId() >= 0;

                        if (p.getDistance() < minDist && isSmallId) {
                            minDist = p.getDistance();
                            personFollow = p;
                        }
                    }

                    if (personFollow != null
                    /* && personFollow.getDistance() < RESUME_MAX_DIST */
                    && noPersonCounter > 5) {
                        try {
                            taskActuator.initiate(new XOPData(builder.build("<"
                                    + PersonFoundNotification.getBaseTag()
                                    + "/>", null)));
                        } catch (ValidityException e) {
                            logger.error(e);
                        } catch (ParsingException e) {
                            logger.error(e);
                        }
                    } else {
                        personFollow = null;
                        // speechActuator.say("I can't see you anymore. "
                        // + "I will follow someone else.");
                        noPersonCounter++;
                    }
                }
            } catch (IOException e) {
                logger.fatal(e.getMessage());
            } catch (InterruptedException e) {
                logger.fatal(e.getMessage());
            }
            personFollowId = -1;
        }
        return personFollow;
    }

    private void follow(PersonData personFollow) throws IOException {

        // logger.info(personFo);
        noPersonCounter = 0;
        logger.debug("follow: Following Person: " + personFollow);

        double personAngle = personFollow.getAngle();
        double personDist = personFollow.getDistance();

        PositionData currentPosition;
        try {

            currentPosition = posSensor.readLast();
            logger.debug("currentPosition: " + currentPosition.getX() + " / "
                    + currentPosition.getY());

            // System.out.println(personDist);

            if (personDist < DISTANCE_THRESHOLD_MOVE
                    && personDist > DISTANCE_THRESHOLD_TURN) {
                logger.debug("personDist< DTM && pD>DTT");
                NavigationGoalData personGoal =
                        CoordinateSystemConverter
                                .polar2NavigationGoalData(
                                        currentPosition,
                                        personAngle,
                                        (personDist - DISTANCE_THRESHOLD_STOP) / 1000.0);

                if (lastKnownPositions.size() >= QUEUE_LENGTH) {
                    lastKnownPositions.poll();
                }
                lastKnownPositions.add(personGoal);

                /*
                 * logger.info("Approaching person [" + personFollow.getId() +
                 * "] (Dist: " + personDist + ")");
                 * 
                 * logger .debug("Person NavigationGoal is x=" +
                 * personGoal.getX() + " y=" + personGoal.getY() + " yaw=" +
                 * personGoal.getYaw());
                 */
                logger.debug("Moving: "
                        + (personDist - DISTANCE_THRESHOLD_STOP) + "m"
                        + "with angle: " + personAngle + " (Radiant)");
                // String currentX = Double.toString(currentPosition.getX())
                // .substring(0, 4);
                // String currentY = Double.toString(currentPosition.getY())
                // .substring(0, 4);
                // String newX = Double.toString(personGoal.getX())
                // .substring(0, 4);
                // String newY = Double.toString(personGoal.getY())
                // .substring(0, 4);
                // logger.debug("Current position: x:" + currentX + " y:"
                // + currentY);
                // logger.debug("New position: x:" + newX + " y:" + newY);

                // personGoal.setYaw(personAngle);

                personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE);
                personGoal.setYawTolerance(YAW_TOLERANCE);
                navActuator.setGoal(personGoal);

            } else if (personDist < DISTANCE_THRESHOLD_TURN) {

                NavigationGoalData personGoal =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                currentPosition, personAngle, 0.0);

                /*
                 * logger .info("Turning towards person [" +
                 * personFollow.getId() + "]");
                 */
                logger.debug("Turning: " + personAngle + " (Radiant)");
                personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE);
                personGoal.setYawTolerance(YAW_TOLERANCE);

                navActuator.setGoal(personGoal);
            }
        } catch (InterruptedException e) {
            logger.fatal("Don't know where I am. " + e.getMessage());
        }
    }

}
