/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util;

import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * Utility class with methods to convert between several coordinate systems.
 */
public class CoordinateSystemConverter {

    /**
     * Generate global carthesian coordinates suitable as a navigation goal for
     * the NavigationActuator from local polar coordinate system coordinates
     * given as distance and angle. This method requires the current global
     * position (OdometryData) of the robot obtained from the OdometrySensor.
     * 
     * 
     * @param positionData
     *            global position of the robot as recently obtained from an
     *            OdometrySensor
     * @param angle
     *            angle of the polar coordinates in radiants (between 0 and 2 *
     *            Math.PI). An angle=0 corresponds to a position directly in
     *            front of the robot. angles are defined counter-clockwise.
     * @param dist
     *            distance of polar coordinates in metres (always >0).
     * @return the global carthesian position in metres and the corresponding
     *         orientation of the robot as NavigationGoalData, which can
     *         directly by used for the NavigationActuator.
     */
    public static NavigationGoalData polar2NavigationGoalData(
            PositionData positionData, double angle, double dist) {
        NavigationGoalData cart = new NavigationGoalData();
        double localAngle = angle + positionData.getYaw();

        if (localAngle > 2 * Math.PI) {
            localAngle = localAngle - 2 * Math.PI;
        }
        if (localAngle < 0) {
            localAngle = localAngle + 2 * Math.PI;
        }
        cart.setX(positionData.getX() + Math.cos(localAngle) * dist);

        cart.setY(positionData.getY() + Math.sin(localAngle) * dist);
        cart.setYaw(localAngle);
        return cart;
    }
    
    public static double positionData2Angle(
            PositionData ownPosition, PositionData objectPosition) {
        
        double angle = Math.atan2(objectPosition.getY() - ownPosition.getY(), objectPosition.getX() - ownPosition.getX());
        
        angle += ownPosition.getYaw();
        
        if (angle >= Math.PI) {
            angle -= 2*Math.PI;
        }
        if (angle < -Math.PI) {
            angle += 2* Math.PI;
        }
        
        return angle;
    }
    
    public static double positionDistance(PositionData ownPosition, PositionData objectPosition) {
        
        double x = ownPosition.getX() - objectPosition.getX();
            
        double y = ownPosition.getY() - objectPosition.getY();
        
        double dist = Math.sqrt( Math.pow(x, 2) + Math.pow(y, 2) );
        
        return dist;
    }

}
