/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.RemoteServer;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.GoalReachedData;
import de.unibi.airobots.btl.data.NavigationGoalData;

/**
 * Actuator which accesses the robots navigation interface. Goals can be
 * provided using NavigationGoalData and are automatically approached using
 * obstacle avoidance.
 * 
 * @author lschilli
 * @author jwienke
 */
public class XcfNavigationActuator implements NavigationActuator,
        RemoteServerReconnectable {

    private static final String DEFAULT_GET_GOAL_METHOD = "getGoal";
    private static final String DEFAULT_MANUAL_STOP_METHOD = "manualStop";
    private static final String DEFAULT_GOAL_REACHED_METHOD = "goalReached";
    private static final String DEFAULT_SET_GOAL_METHOD = "setGoal";

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";
    private RemoteServer remoteServer = null;
    private Object remoteServerLock = new Object();

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server to connect to
     */
    public XcfNavigationActuator(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setGoal(NavigationGoalData data) throws IOException {
        try {

            Document doc =
                    new Document(new Element(NavigationGoalData.getBaseTag()));
            data.fillInto(doc.getRootElement());

            logger.debug("Calling " + DEFAULT_SET_GOAL_METHOD + " with"
                    + doc.toXML());

            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_SET_GOAL_METHOD, new XOPData(
                        doc));
            }

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public GoalReachedData isGoalReached() throws IOException {
        try {
            logger.debug("Calling " + DEFAULT_GOAL_REACHED_METHOD);
            XOPData xop;
            synchronized (remoteServerLock) {
                xop =
                        remoteServer.callMethod(DEFAULT_GOAL_REACHED_METHOD,
                                new XOPData());
            }
            GoalReachedData grd = new GoalReachedData();
            GoalReachedData
                    .fromElement(xop.getDocument().getRootElement(), grd);
            return grd;
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void manualStop() throws IOException {
        try {
            logger.debug("Calling " + DEFAULT_MANUAL_STOP_METHOD);
            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_MANUAL_STOP_METHOD,
                        new XOPData());
            }
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHEKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public NavigationGoalData getCurrentGoal() throws IOException {
        try {

            XOPData xop;
            synchronized (remoteServerLock) {
                xop =
                        remoteServer.callMethod(DEFAULT_GET_GOAL_METHOD,
                                new XOPData());
            }
            NavigationGoalData goal = new NavigationGoalData();
            NavigationGoalData.fromElement(xop.getDocument().getRootElement(),
                    goal);
            return goal;

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

}
