/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.yarp;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;

import org.apache.log4j.Logger;
import yarp.Contactable;
import yarp.Network;


/**
 * Manages (initilizes and re-connects) {@link PortReconnectable} opbjects.
 *
 * @author cdondrup
 */
public class PortReconnectableManager<P extends Contactable>
        extends ReconnectableManager<PortReconnectable<P>> {

    private Object instance;

    private Logger logger = Logger.getLogger(getClass());
    private Object portLock = new Object();
    private Map<String, PortWrapper<P>> portsByName =
            new HashMap<String, PortWrapper<P>>();
    private Map<PortReconnectable<P>, PortWrapper<P>> portsByReconnectable =
            new HashMap<PortReconnectable<P>, PortWrapper<P>>();
    private Timer checkTimer = new Timer("port check timer");
    private PortChecker portCheckerTask = new PortChecker();


    /**
     * Checks all ports if they are connected or not
     */
    private class PortChecker extends TimerTask {

        /**
         * Synchronized check for all ports if they are connected or not.
         * Tries to connect the ports if they are not.
         */
        @Override
        public void run() {

            logger.debug("Checking ports starts...");

            synchronized (portLock) {

                // iterate over all managed reconnectables
                for (Entry<PortReconnectable<P>, PortWrapper<P>> portEntry : portsByReconnectable.entrySet()) {

                    String portName = portEntry.getKey().getPortName();

                    try {

                        if (portEntry.getValue() == null) {
                            logger.debug("Port '" + portName + "' does not exist.");
                            return;
                        }

                        PortWrapper<P> checkPort =
                                portEntry.getValue();
                        PortReconnectable<P> reconnectable =
                                portEntry.getKey();

                        if (Network.isConnected(portName, checkPort.getConnectTo())
                                || Network.isConnected(checkPort.getConnectTo(), portName)) {

                            logger.debug("Port '" + portName
                                    + "' seems to be alive.");
                        } else {
                            logger.debug("Port '" + portName
                                    + "' seems to be dead."
                                    + " Trying to reconnect.");
                            reconnectable.resetPort(checkPort);

                        }




                        // CHECKSTYLE:OFF catch everything here to not break the
                        // checking thread
                    } catch (Throwable e) {
                        // CHECKSTYLE:ON

                        // connection to subscriber failed, mark it as
                        // unconnected

                        logger.warn("Port with name '" + portName
                                + "' could not be connected "
                                + "or reinitialized. ");
                        logger.debug(e);



                    }

                }

            }

            logger.debug("Checking ports finished");

        }
    }

    /**
     * Standard constructor using the default check interval time of 5 seconds
     */
    public PortReconnectableManager() {
        this(DEFAULT_CHECK_INTERVAL_IN_MS);
    }

    /**
     * Constructor with custom check interval time
     *
     * @param checkIntervalMs The check interval in milliseconds
     */
    public PortReconnectableManager(long checkIntervalMs) {
        checkTimer.scheduleAtFixedRate(portCheckerTask, checkIntervalMs, checkIntervalMs);
    }

    /**
     * Creates new instances for all ports in portMap.
     * 
     * @param portMap Map of two Strings port name and the name of the port to connect to
     */
    @Override
    public void initializeYarpPorts(Map<String, String> portMap) {

        logger.info("Initializing ports: ." + portMap.keySet());

        for (String portName : portMap.keySet()) {
            if (!portsByName.containsKey(portName)) {
                createPort(portName, portMap.get(portName));
            }
        }
    }

    /**
     * Creates a new port with given name and target.
     * Called by initializeYarpPorts.
     * 
     * @param portName Port name
     * @param connectTo Name of port to connect to
     */
    public void createPort(String portName, String connectTo) {

        synchronized (portLock) {
            try {
                portsByName.put(portName,
                        new PortWrapper<P>((P) instance,
                        connectTo));
                portsByName.get(portName).getPort().open(portName);
            } catch (Exception e) {
                logger.debug(e);
            }
        }
    }

    /**
     * Initializes and manages the given {@link PortReconnectable} for the lifetime
     * of this class.
     * 
     * @param cls The Class of the port to manage
     * @param reconnectable The {@link PortReconnectable} to manage
     */
    @Override
    public void initializeAndManage(Class cls, PortReconnectable<P> reconnectable) {

        try {
            instance = cls.newInstance();
        } catch (InstantiationException ex) {
            java.util.logging.Logger.getLogger(PortReconnectableManager.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IllegalAccessException ex) {
            java.util.logging.Logger.getLogger(PortReconnectableManager.class.getName()).log(Level.SEVERE, null, ex);
        }
        String portName = reconnectable.getPortName();

        logger.info("Got request to initialize and manage new reconnectable: "
                + reconnectable + " for port with name '"
                + portName + "'.");

        synchronized (portLock) {
            if (!portsByName.containsKey(portName)) {
                initializeYarpPort(portName, reconnectable.getConnectTo());

                logger.warn("There was no port for name '"
                        + portName
                        + "' initialized. Creation was needed.");
            }

            assert portsByName.containsKey(portName);

            if (portsByName.get(portName) != null) {
                PortWrapper<P> portWrapper = portsByName.get(portName);
                logger.info("Initially setting port " + portName
                        + " for " + reconnectable);
                reconnectable.resetPort(portWrapper);
                if (!portsByReconnectable.containsKey(reconnectable)) {
                    portsByReconnectable.put(reconnectable,
                            portsByName.get(portName));
                } else {
                    logger.error("The port '" + portName
                            + "' allready exists in '" + reconnectable
                            + "'. Could not be added to list.");
                }
            } else {
                logger.warn("There was no current connection "
                        + "to a port with name '" + portName
                        + "'. Initialization of " + reconnectable
                        + " postponed.");

            }

        }
    }

    public PortWrapper<P> getPortByName(String portName) {
        return portsByName.get(portName);
    }
}
